#include <stdio.h>
#include <stdlib.h>
#include <string.h>

static char *skip_ws( char *text ) {
  while ( *text == ' ' ) text++;
  return text;
}

static int get_reg_code( char*& text ) {
  char *reg_name[] = {
	"r0", "r1",  "r2",  "r3",  "r4",  "r5",  "r6",  "r7",
	"r8", "r9", "r10", "r11", "r12", "r13", "r14", "r15"
  };
  int match_len = -1, match_idx = -1;
  int regname_len;

  for ( int i = 0; i < 16; i++ ) {
  	regname_len = strlen(reg_name[i]);
	if ( !strncasecmp( text, reg_name[i], regname_len ) &&
	     match_len < regname_len ) {

		match_idx = i;
		match_len = strlen(reg_name[i]);
	}
  }

  if ( match_idx != -1 ) {
	text += match_len;
	return match_idx;
  }

  if ( !strncasecmp( text, "ip", 2 ) ) {
  	text += 2;
	return 0x0c;
  }
  else if ( !strncasecmp( text, "sp", 2 ) ) {
  	text += 2;
	return 0x0d;
  }
  else if ( !strncasecmp( text, "lr", 2 ) ) {
  	text += 2;
	return 0x0e;
  }
  else if ( !strncasecmp( text, "pc", 2 ) ) {
  	text += 2;
	return 0x0f;
  }

  return -1;
}

static int get_imm_code( char*& text ) {
  unsigned val;
  char *s, *ep;

  s = text;

  if ( *s != '#' ) return -1;
  *s++;

  if ( *s == '&' ) {
	s++;
	val = strtoul( s, &ep, 16 );
  }
  else
	val = strtoul( s, &ep, 10 );

  if ( s == ep ) return -1;

  text = ep;
  return val;
}

static int get_imm_code( char*& text, int& ror_value ) {
  unsigned val;
  char *s, *ep;

  s = text;

  val = get_imm_code( s );
  if ( val == -1 ) return -1;

  for ( ror_value = 0; ror_value < 16; ror_value++ ) {
	if ( val < 256 ) break;
	val = (val << 2) | (val >> 30);
  }
  if ( ror_value == 16 ) return -1;

  text = s;

  return val;
}

static int get_shift_code( char*& text, int& rc_value ) {
  char *shift_op[] = {
  	"lsl", "lsr", "asr", "ror"
  };

  int i, shift_code;
  char *s = text, *ep;

  for ( i = 0; i < 4; i++ )
	if ( !strncasecmp( s, shift_op[i], 3 ) ) {
		s += 3;
		break;
	}

  if ( !strncasecmp( s, "asl", 3 ) ) {
  	i = 0;
  	s += 3;
  }

  if ( i == 4 || *s != ' ' ) return -1;

  s = skip_ws( s );
  shift_code = i * 2;

  if ( *s == '#' ) {
  	s++;
  	rc_value = strtoul( s, &ep, 10 );
	if ( ep == s ) return -1;

	s = ep;
  }
  else {
  	rc_value = get_reg_code( s );
	if ( rc_value == -1 ) return -1;
	rc_value <<= 1;
	shift_code++;
  }

  text = s;

  return shift_code;
}

static int get_cond_code( char*& text ) {
  char *cond[] = {
	"eq", "ne", "cs", "cc", "mi", "pl", "vs", "vc",
	"hi", "ls", "ge", "lt", "gt", "le", "al", "nv"
  };

  for ( int i = 0; i < 16; i++ )
	if ( !strncasecmp( text, cond[i], 2 ) ) {
		text += 2;
		return i;
	}

  return -1;
}

#define INSTR_SETS 1
#define INSTR_HASP 2
#define INSTR_NORD 4

struct instr_desc {
  const char *name;
  int argnum, opt;
};

static bool process_data_instr( char *text, char*& code ) {
  instr_desc instr[] = {
	{"and", 3},
	{"eor", 3},
	{"sub", 3},
	{"rsb", 3},
	{"add", 3},
	{"adc", 3},
	{"sbc", 3},
	{"rsc", 3},
	{"tst", 2, INSTR_SETS | INSTR_HASP | INSTR_NORD },
	{"teq", 2, INSTR_SETS | INSTR_HASP | INSTR_NORD },
	{"cmp", 2, INSTR_SETS | INSTR_HASP | INSTR_NORD },
	{"cmn", 2, INSTR_SETS | INSTR_HASP | INSTR_NORD },
	{"orr", 3},
	{"mov", 2},
	{"bic", 3},
	{"mvn", 2}
  };

  char *s = skip_ws( text );
  int i, val, argnum, opt, instr_code = -1, cond_code = -1;
  int rd_code, rn_code, rm_code, rc_code;
  int imm_code, ror_code, shift_code;
  bool has_sflag = false, has_pflag = false;
  bool has_imm_code = false, has_shift_code = false;

  for ( i = 0; i < 15; i++ )
	if ( !strncasecmp( s, instr[i].name, 3 ) ) {
		instr_code = i;
		argnum = instr[i].argnum;
		opt = instr[i].opt;
		if ( opt & INSTR_SETS ) has_sflag = true;
		break;
	}

  if ( instr_code == -1 ) return false;

  s += 3;

  cond_code = get_cond_code( s );
  if ( cond_code == -1 ) cond_code = 0x0e;

  if ( *s == 's' ) {
  	has_sflag = true;
	s++;
  }

  if ( opt & INSTR_HASP && *s == 'p' ) {
  	has_pflag = true;
	s++;
  }

  if ( *s != ' ' ) return false;

  s = skip_ws( s );

  rd_code = get_reg_code( s );
  if ( rd_code == -1 ) return false;
  s = skip_ws( s );
  if ( *s != ',' ) return false;
  s = skip_ws( s + 1 );

  if ( argnum == 3 ) {
	rn_code = get_reg_code( s );
	if ( rn_code == -1 ) return false;
	s = skip_ws( s );
	if ( *s != ',' ) return false;
	s = skip_ws( s + 1 );
  }
  else {
  	if ( opt & INSTR_NORD ) {
	  	rn_code = rd_code;
		rd_code = 0;
	}
	else
	  	rn_code = 0;
  }
  if ( has_pflag ) rd_code = 0x0f;

  rm_code = get_reg_code( s );
  if ( rm_code == -1 ) {
	imm_code = get_imm_code( s, ror_code );
	if ( imm_code == -1 ) return false;

	has_imm_code = true;
  }
  else {
  	s = skip_ws( s );
	if ( *s == ',' ) {
		s = skip_ws( s + 1 );

		shift_code = get_shift_code( s, rc_code );
		if ( shift_code == -1 ) return false;

		has_shift_code = true;
	}
  }

  s = skip_ws( s );
  if ( *s  ) return false;

  if ( has_imm_code )
	val = imm_code;
  else {
	val = rm_code;
	if ( has_shift_code ) 
		val |= (rc_code << 7) | (shift_code << 4);
  }

  *code++ = val & 0xff;

  val = (rd_code << 4 );
  if ( has_imm_code )
  	val |= ror_code;
  else if ( has_shift_code )
  	val |= rc_code >> 1;
  
  *code++ = val & 0xff;

  val = (instr_code << 5) | rn_code;
  if ( has_sflag ) val |= 0x10;

  *code++ = val & 0xff;

  val = (cond_code << 4) | (instr_code >> 3);
  if ( has_imm_code ) val |= 2;

  *code++ = val & 0xff;

  return true;
}

struct block_instr_sfx {
  char *sfx, *lsfx, *ssfx;
  int pflag, uflag;
};

static int get_reg_interval_code( char*& text ) {
  int reg_code1, reg_code2, regint;
  char *s = text;

  reg_code1 = get_reg_code( s );
  if ( reg_code1 == -1 ) return -1;

  s = skip_ws( s );
  if ( *s != '-' ) {
  	text = s;
  	return 1 << reg_code1;
  }

  s++;
  s = skip_ws( s );

  reg_code2 = get_reg_code( s );
  if ( reg_code2 == -1 || reg_code2 <= reg_code1 ) return -1;

  regint = 0;
  while ( reg_code1 <= reg_code2 ) {
  	regint |= 1 << reg_code1;
	reg_code1++;
  }

  text = s;
  return regint;
}

static int get_regblock_code( char*& text ) {
  int blockcode = 0, intcode;
  char *s = text;

  if ( *s != '{' ) return -1;

  s++;
  s = skip_ws( s );

  for (;;) {
  	intcode = get_reg_interval_code( s );
	if ( intcode == -1 ) return -1;

	blockcode |= intcode;

	s = skip_ws( s );
	if ( *s == '}' ) {
		s++;
		break;
	}
	if ( *s != ',' ) return -1;

	s++;
	s = skip_ws( s );
  }

  text = s;
  return blockcode;
}

static bool process_block_data_instr( char *text, char*& code ) {
  int i, cond_code = 0x0e, base_reg, reg_code, regmap = 0, val;
  int lflag, pflag, uflag, wflag = 0, sflag = 0;
  char *s;

  block_instr_sfx instr_sfx[] = {
  	{ "da", "fa", "ed", 0, 0 },
	{ "db", "ea", "fd", 1, 0 },
	{ "ia", "fd", "ea", 0, 1 },
	{ "ib", "ed", "fa", 1, 1 }
  };

  s = text;

  if ( !strncasecmp( s, "stm", 3 ) ) {
  	s += 3;
  	lflag = 0;
  }
  else if ( !strncasecmp( s, "ldm", 3) ) {
  	s += 3;
  	lflag = 1;
  }
  else return false;

  for ( i = 0; i < 4; i++ )
  	if ( !strncasecmp( s, instr_sfx[i].sfx, 2 ) ||
	     lflag && !strncasecmp( s, instr_sfx[i].lsfx, 2 ) ||
	     !lflag && !strncasecmp( s, instr_sfx[i].ssfx, 2 ) ) {

		s += 2;
		pflag = instr_sfx[i].pflag;
		uflag = instr_sfx[i].uflag;
		break;
	}
  if ( i == 4 ) return false;

  if ( *s != ' ' ) return false;
  s = skip_ws( s );

  base_reg = get_reg_code( s );
  if ( base_reg == -1 ) return false;

  if ( *s == '!' ) {
  	s++;
	wflag = 1;
  }

  s = skip_ws( s );
  if ( *s != ','  ) return false;

  s++;
  s = skip_ws( s );

  reg_code = get_regblock_code( s );
  if ( reg_code == -1 ) return false;

  s = skip_ws( s );
  if ( *s == '^' ) {
  	sflag = true;
	s++;
	s = skip_ws( s );
  }

  if ( *s ) return false;

  val = reg_code & 0xff;
  *code++ = val;

  val = reg_code >> 8;
  *code++ = val;

  val = (uflag << 7) | (sflag << 6) | (wflag << 5) | (lflag << 4);
  val |= base_reg;
  *code++ = val;

  val = (cond_code << 4) | 8 | pflag;
  *code++ = val;

  return true;
}

int current_ip;

static int get_addr_ref( char*& text ) {
  unsigned val;
  char *s, *ep;

  s = text;

  if ( *s != '&' ) return -1;
  s++;

  val = strtoul( s, &ep, 16 );
  if ( s == ep ) return -1;

  text = ep;
  return val;
}

static bool process_data_transfer_instr( char *text, char*& code ) {
  int val, cond_code = 0x0e;
  int uflag = 1, pflag = 1, lflag, bflag = 0, wflag = 0;
  int dest_reg, base_reg, offs_reg, shift_code, rc_code, offs, addr;
  bool has_imm_offs = false;
  char *s = text;

  if ( !strncasecmp( text, "ldr", 3 ) ) {
  	s += 3;
  	lflag = 1;
  }
  else if ( !strncasecmp( text, "str", 3 ) ) {
  	s += 3;
  	lflag = 0;
  }
  else return false;

  if ( *s == 'b' ) {
  	s++;
	bflag = 1;
  }

  if ( *s == 't' ) {
  	s++;
  	wflag = 1;
  }
  
  s = skip_ws( s );

  dest_reg = get_reg_code( s );
  if ( dest_reg == -1 ) return false;

  s = skip_ws( s );
  if ( *s != ',' ) return false;

  s++;
  s = skip_ws( s );

  addr = get_addr_ref( s );
  if ( addr != -1 ) {
  	offs = addr - current_ip - 8;
	if ( offs < 0 ) {
		uflag = 0;
		offs = -offs;
	}
  	if ( offs > 0xfff ) return false;

	base_reg = 0x0f;
	has_imm_offs = true;

  	s = skip_ws( s );
  	goto finish_parsing;
  }

  if ( *s != '[' ) return false;

  s++;
  s = skip_ws( s );

  base_reg = get_reg_code( s );
  if ( base_reg == -1 ) return false;

  s = skip_ws( s );
  if ( *s == ']' ) {
  	pflag = 0;
	s++;
	s = skip_ws( s );
  }
  else
	if ( wflag ) return false;

  if ( *s == ',' ) {
	s++;
	s = skip_ws( s );

	offs_reg = get_reg_code( s );
	if ( offs_reg != -1 ) {
		s = skip_ws( s );
		if ( *s == ',' ) {
			shift_code = get_shift_code( s, rc_code );
			if ( shift_code == -1 || (shift_code & 1) )
				return false;
		}
	}
	else {
		offs = get_imm_code( s );
		if ( offs == -1 || offs > 0xfff ) return false;
		has_imm_offs = true;
	}
  }
  else {
	offs = 0;
  	has_imm_offs = true;
  }

  s = skip_ws( s );

  if ( pflag ) {
	if ( *s != ']' ) return false;
	s++;
	s = skip_ws( s );
  }

  if ( *s == '!' ) {
  	if ( wflag ) return false;
	wflag = 1;
	s++;
	s = skip_ws( s );
  }

finish_parsing:
  if ( *s ) return false;

  if ( has_imm_offs ) val = offs & 0xff;
  else 
  	val = (rc_code << 7) | (shift_code << 4) | offs_reg;

  *code++ = val;

  val = dest_reg << 4;
  if ( has_imm_offs ) val |= offs >> 8;
  else val |= rc_code >> 1;

  *code++ = val;

  val = (uflag << 7) | (bflag << 6) | (wflag << 5) | (lflag << 4) | base_reg;
  *code++ = val;

  val = (cond_code << 4) | 4 | pflag;
  if ( !has_imm_offs ) val |= 2;

  *code++ = val;

  return true;
}

static bool process_branch_instr( char *text, char*& code ) {
  int val, lflag = 0, cond_code, addr;
  char *s = text;

  if ( *s != 'b' ) return false;
  s++;

  cond_code = get_cond_code( s );
  if ( cond_code == -1 ) {
	if ( *s == 'l' ) {
	  	s++;
	  	lflag = 1;
		cond_code = get_cond_code( s );
	}
	if ( cond_code == -1 ) cond_code = 0x0e;
  }

  if ( *s != ' ' ) return false;
  s = skip_ws( s );

  addr = get_addr_ref( s );
  if ( addr == -1 ) return false;

  addr = (addr - current_ip - 8) / 4;

  *code++ = addr & 0xff;
  *code++ = (addr >> 8) & 0xff;
  *code++ = (addr >> 16) & 0xff;
  val = (cond_code << 4) | 0x0a | lflag;
  *code++ = val;


  return true;
}

char asmbuf_arm[10], *asmbuf_ptr_arm;

bool do_asm_str_arm( char *str ) {

  asmbuf_ptr_arm = asmbuf_arm;

  if ( process_data_instr( str, asmbuf_ptr_arm ) ||
       process_data_transfer_instr( str, asmbuf_ptr_arm ) ||
       process_block_data_instr( str, asmbuf_ptr_arm ) ||
       process_branch_instr( str, asmbuf_ptr_arm) ) return true;

  return false;
}

#ifdef MAIN

int read_be_int( char *data ) {
  return ((data[0] & 0xff) << 24) | ((data[1] & 0xff) << 16) |
  	((data[2] & 0xff) << 8 ) | (data[3] & 0xff);
}

int main() {
  char code_buf[200], *s, *p;
  char *text0[] = {
  	"mov r1, #0",
	"addeqs r2, r1, r4, asl r3",
	"andeq r1, r5, r2",
	"teqp r4, #&80000000",
	"cmp r5, r1",
	0
  };
  char *text[] = {
#if 0
  	"stmfd sp!, {r4-r11,lr}",
	"sub sp, sp, #16",
	"mov r8, r0",
	"ldr r0, &00030300",
	"mov r7, r2",
	"ldr r2, [sp, #60]",
	"ldr r9, [sp, #64]",
	"ldr r10, [sp, #56]",
	"ldr r11, [sp, #52]",
	"ldr ip, [r0, #0]",
	"mov r4, r1",
	"mov r5, r3",
	"mov r3, #&82",
	"cmp ip, #0",
	"ldr r1, &00030304",
	"ldr r0, &0002fda0",
	"mov r6, #0",
	"beq &0003004c",
	"cmp r9, #0",
	"bne &0003004c",
#endif
	"cmp r2, #0",
	"bne &0002ff6c",
	"ldrb r2, [r5, #0]",
	"cmp r2, #0",
	"beq &000302f4",
	"ldrb r0, [r5, #0]",
	"cmp r0, #&80",
	"blt &0002feec",
	"ldr r0, &0002fda0",
	"ldrb r1, [r0, #8]",
	"cmp r1, #0",
	"beq &0002fe38",
	"ldrb r0, [r0, #4]",
  	0
  };

  p = code_buf;
//  current_ip = 0x2fda4;
  current_ip = 0x2fdf4;
  for ( int i = 0; text[i]; i++ ) {
  	printf( "%s", text[i] );

	if ( do_asm_str_arm( text[i] ) ) {
		puts( " - ok" );

		memmove( p, asmbuf_arm, asmbuf_ptr_arm - asmbuf_arm );
		p += asmbuf_ptr_arm - asmbuf_arm;
	}
	else
		puts( " - failed" );
	
	current_ip += 4;
  }

  s = code_buf;
  while ( s < p ) {
  	printf( "%08x\n", read_be_int( s ) );
	s += 4;
  }

  return 0;
}

#endif // MAIN
